'''A simple logger'''

import os
import datetime
import uuid

def gen_uuid() -> str:
    '''
        Generates a UUID
    '''
    return str(uuid.uuid4())


class LoggerError(Exception):
    '''Exception raised for log file related errors'''


class Logger:
    '''A simple logger class'''

    __slots__: list[str] = ["file_path",
                            "encoding",
                            "time_format",
                            "separator",
                            "__silent",
                            "__enable_remove",
                            "__strip_content"]

    def __init__(self,
                 file_path: str | None = None,
                 encoding: str = "utf-8-sig",
                 time_format: str = "%Y.%m.%d %H:%M:%S",
                 separator: str = "\t",
                 silent: bool = False,
                 enable_remove: bool = False,
                 strip_content: bool = False) -> None:
        '''
            Logger class

            .. code-block:: python
                demo_logger = Logger(file_path = "example.log",
                                     strip_content = True)

            :param file_path: :class:`Optional(Union(str, None))` File's path. Defaults to `None`
            :param encoding: :class:`Optional(str)` File's encoding. Defaults to `"utf-8-sig"`
            :param time_format: :class:`Optional(str)` Log's time format. Defaults to `"%Y.%m.%d %H:%M:%S"`
            :param spearator: :class:`Optional(str)` Log's separator betwen the time and the content. Defaults to `"\\t"`
            :param silent: :class:`Optional(bool)` If set to `True` then won't print to console. Defaults to `False`
            :param enable_remove: :class:`Optional(bool)` If set to `True` then it enables removing content. Defaults to `False`
            :param strip_content: :class:`Optional(bool)` If set to `True` then strips content before logging it. Defaults to `False`
        ''' # pylint: disable=line-too-long
        self.file_path: str = file_path if file_path else os.path.join(os.getcwd(),
                                                                       f"{gen_uuid()}.log")
        self.encoding: str = encoding
        self.time_format: str = time_format
        self.separator: str = separator
        self.__silent: bool = silent
        self.__enable_remove: bool = enable_remove
        self.__strip_content: bool = strip_content


    def __str__(self) -> str:
        return f"Log file: {self.file_path}"


    def __error(self,
                message: str | None = None) -> LoggerError:
        '''
            Prints error message

            Parameters:
                message (str): error message
        '''
        raise LoggerError(str(message or ""))


    def __str_time(self) -> str:
        '''
            Returns the current time as a string
        '''
        current_time = datetime.datetime.now()
        try:
            return current_time.strftime(self.time_format)
        except Exception as e: # pylint: disable=broad-except
            print(f"Error: {e}")
            return current_time.strftime("%Y.%m.%d %H:%M:%S")


    def __log_to_file(self,
                      content: any) -> None:
        '''
            Appends file

            Args:
                content (str): content
        '''
        try:
            with open(file = self.file_path,
                      mode = "a+",
                      encoding = self.encoding) as file:
                file.write(str(content))
        except Exception as e:
            raise LoggerError(f"Error writing to file: {self.file_path}, error: {e}") from e


    def __strip(self,
                content: str) -> str:
        '''
            Strips the content

            Parameters:
                content (str): content
        '''
        return content.strip() if self.__strip_content else content


    def log(self,
            content: any = None) -> None:
        '''
            Logs content to file

            .. code-block:: python
                demo_logger.log(content = "Example")

            :param content: :class:`any` Content to log
        '''
        self.__log(content = content)


    def change_path(self,
                    file_path: str) -> None:
        '''
            Changes the log file path

            .. code-block:: python
                demo_logger.change_path(file_path = "example.log")

            :param file_path: :class:`str` New path to the log file
        '''
        self.__change_path(file_path = file_path)


    def change_encoding(self,
                        encoding: str) -> None:
        '''
            Changes the log file encoding

            .. code-block:: python
                demo_logger.change_encoding(encoding = "utf-8-sig")

            :param encoding: :class:`str` New encoding to the log file
        '''
        self.__change_encoding(encoding = encoding)


    def change_time_format(self,
                           time_format: str) -> None:
        '''
            Changes the time format

            .. code-block:: python
                demo_logger.change_time_format(time_format = "%Y.%m.%d %H:%M:%S")

            :param time_format: :class:`str` New time format for the log file
        '''
        self.__change_time_format(time_format = time_format)


    def change_separator(self,
                         separator: str) -> None:
        '''
            Changes the separator
            
            .. code-block:: python
                demo_logger.change_separator(separator = "\\t")

            :param separator: :class:`str` New separator for the log file
        '''
        self.__change_separator(separator = separator)


    def clear(self) -> None:
        '''
            Clears the log file
        '''
        self.__clear()


    def remove(self) -> None:
        '''
            Removes the log file
        '''
        self.__remove()


    def __log(self,
              content: str = "") -> None:
        content = self.__str_time() + self.separator + str(self.__strip(content)) + "\n"
        if not self.__silent:
            print(content.strip())
        self.__log_to_file(content)


    def __change_path(self,
                      file_path: str) -> None:
        self.file_path = file_path
        print(f"Changed path from {self.file_path} to {file_path}")


    def __change_encoding(self,
                          encoding: str) -> None:
        self.encoding = encoding
        print(f"Changed encoding to {encoding}")


    def __change_time_format(self,
                             time_format: str) -> None:
        self.time_format = time_format
        print(f"Changed time format to {time_format}")


    def __change_separator(self,
                           separator: str) -> None:
        self.separator = separator
        print(f"Changed separator to {separator}")


    def __clear(self) -> None:
        if self.__enable_remove:
            if os.path.exists(self.file_path):
                with open(self.file_path,
                          "w",
                          encoding = self.encoding) as _:
                    print(f"Log file cleared ({self.file_path})")
            self.__error(f"Log file does not exist ({self.file_path})")
        self.__error("Removal is not enabled")


    def __remove(self) -> None:
        if self.__enable_remove:
            if os.path.exists(self.file_path):
                os.remove(self.file_path)
                print(f"Log file removed ({self.file_path})")
            self.__error(f"Log file does not exist ({self.file_path})")
        self.__error("Removal is not enabled")



'''
    Excel writer module
'''

import os
from datetime import date
from decimal import Decimal
import uuid
import xlsxwriter

def gen_uuid(length: int = 8) -> str:
    '''
        Generating a random UUID
        
        .. code-block:: python
            uuid_str: str = gen_uuid(length = 16)

        :param length: :class:`int` length of the returned uuid (default: 8)
    '''
    return str(uuid.uuid4())[:length]


def is_list(obj: any) -> bool:
    '''
        Return true if the object is a list else false
        
        .. code-block:: python
            list_bool: bool = is_list(variable)

        :param obj: :class:`any` object to check
    '''
    return isinstance(obj, list)


class WorkSheet:
    '''
        Worksheet class
    '''

    MIN_WIDTH: int = 10
    MAX_WIDTH: int = 40

    __slots__: list[str] = ["name",
                            "header",
                            "data",
                            "header_comment"]

    def __init__(self,
                 name: str,
                 header: list[str] = None,
                 data: list[list[any]] = None,
                 header_comment: list = None) -> None:
        '''
            Worksheet class which later on need to be passed to the Workbook class.

            .. code-block:: python

                sheet = WorkSheet(name = "sheet_1",
                                  header = ["column_1", "column_2"],
                                  data = [["data_1_1", "data_1_2"],
                                          ["data_2_1", "data_2_2"]])

            :param name: :class:`str` name of the worksheet.
            :param header: :class:`list[str]` or :class:`None` (list[str] or None) Header of the worksheet.
            :param data: :class:`list[list[str]]` or :class:`None` Data matrix of the worksheet
            :param header_comment: :class:`list[list[str]]` or :class:`None` Comment for header

        ''' # pylint: disable=line-too-long
        self.name: str = name if len(name) < 32 else name[:31]
        self.header: list[str] = header
        self.data: list[list] = data
        self.header_comment: list = header_comment if header_comment else []


    # Seggesting a column width with min. and max. filter
    def suggest_width(self,
                      col: int,
                      min_width: int = MIN_WIDTH,
                      max_width: int = MAX_WIDTH) -> None:
        '''
            Suggesting a column width with min. and max. filter
        
            :param col: :class:`int` Column index
            :param min_width: :class:`int` Minimum width. Defaults to `MIN_WITDH` variable.
            :param max_width: :class:`int` Maximum width. Defaults to `MAX_WIDTH` variable
        '''

        length: int = min_width
        if self.header:
            if len(str(self.header[col])) > length:
                length: int = len(str(self.header[col]))
        if self.data:
            for row in self.data:
                if len(row) > col and len(str(row[col])) > length:
                    length: int = len(str(row[col]))

        return (length if length < max_width else max_width)


    def set_min_width(self,
                      min_width: int) -> None:
        '''
            Setting the minimum width
        
            :param min_width: :class:`int` Minimum width
        '''
        self.MIN_WIDTH = min_width


    def set_max_width(self,
                      max_width: int) -> None:
        '''
            Setting the maximum width
        
            :param max_width: :class:`int` Maximum width
        '''
        self.MAX_WIDTH = max_width


class WorkBook:
    '''Workbook class'''

    __slots__: list[str] = ["name",
                            "sheets",
                            "__is_list",
                            "sheet_count",
                            "__uuid"]

    def __init__(self,
                 name: str,
                 sheets: list[WorkSheet] | WorkSheet) -> None:
        '''
            Workbook class

            .. code-block:: python
                book = Workbook(name = "book_1",
                                sheets = [sheet_1, sheet_2])

            :param name: :class:`str` Name of the workbook
            :param sheets: :class:`Union(list[WorkSheet], WorkSheet)` worksheets to fill with
        '''
        self.name: str = name
        self.sheets: list[WorkSheet] = sheets
        self.__is_list: bool = is_list(sheets)
        self.sheet_count: int = self.__get_sheet_count()
        self.__uuid: str = gen_uuid()


    def __get_sheet_count(self) -> int:
        '''
            Return the number of sheets
        '''
        return 1 if not self.__is_list else len(self.sheets)


    # Creating the .xlsx
    def xlsx_create(self,
                    file_path: str = None) -> str:
        '''
            Creating the .xlsx

            .. code-block:: python
                book.xlsx_create(file_path = "example.xlsx")

            :param file_path: :class:`str` Excel create path
        '''

        file_path = file_path or os.path.join(os.getcwd(),
                                              f"{self.__uuid}.xlsx")

        # Creating the .xlsx
        file: xlsxwriter.Workbook = xlsxwriter.Workbook(filename = file_path)
        bold_format = file.add_format({"bold": True})
        date_format = file.add_format({'num_format': 'yyyy.mm.dd'})
        number_format = file.add_format({'num_format': '#,##0.00'})

        # Fetching the worksheet(s)
        worksheets: list = []
        if self.__is_list:
            for sheet in self.sheets:
                worksheets.append(sheet)
        else:
            worksheets.append(self.sheets)

        print(f"{file_path} is {str(len(worksheets))} worksheet{'s' if self.__is_list else ''}:")

        # Creating the worksheet(s)
        for wsheet in worksheets:
            print(f"\t {wsheet.name}")
            sheet = file.add_worksheet(wsheet.name)
            row: int = 0
            col: int = 0
            last_row: int = row
            last_col: int = col

            # If available, then creating a header
            if wsheet.header:
                sheet.freeze_panes(1, 0)
                for element in wsheet.header:
                    width = wsheet.suggest_width(col)
                    sheet.set_column(col, col, width)
                    sheet.write(row, col, element, bold_format)
                    # If available, then adding comments
                    if wsheet.header_comment:
                        for comment in wsheet.header_comment:
                            if element == comment[0]:
                                sheet.write_comment(row, col, comment[1])
                    col += 1
                    last_col = (col if col > last_col else last_col)
                row += 1

            # If available, then creating the data
            if wsheet.data:
                last_row = (row if row > last_row else last_row)
                for line in wsheet.data:
                    col = 0
                    for element in line:
                        # Checking for basic types
                        if isinstance(element, date):
                            try:
                                sheet.write_datetime(row, col, element, date_format)
                            except Exception: # pylint: disable=broad-exception-caught
                                pass
                        elif isinstance(element, Decimal):
                            try:
                                sheet.write(row, col, element, number_format)
                            except Exception: # pylint: disable=broad-exception-caught
                                pass
                        else:
                            try:
                                sheet.write(row, col, element)
                            except Exception: # pylint: disable=broad-exception-caught
                                pass
                        col += 1
                        last_col = (col if col > last_col else last_col)
                    row += 1
                    last_row = (row if row > last_row else last_row)

            # If header is available, then adding an aoutfilter
            if wsheet.header:
                if last_col != 0:
                    last_col -= 1
                sheet.autofilter(0, 0, last_row, last_col)

        # Closing the .xlsx
        file.close()

        print(f"xlsx created: {file_path}")

        return file_path



'''
    Excel reader module
'''

import pandas
import numpy

class ReadExcel:
    '''
        Excel reader class
    '''
    __slots__: list[str] = ["path",
                            "data"]

    def __init__(self,
                 path: str,
                 read_on_init: bool = False) -> None:
        '''
            Excel reader class

            .. code-block:: python
                xlsx = ReadExcel(path = "example.xlsx",
                                 read_on_init = True)

            :param path: :class:`str` Path to the excel
            :param read_on_init: :class:`Optional(bool)` Read excel's content on init. Defaults to `False`
        ''' # pylint: disable=line-too-long
        self.path: str = path
        self.data: pandas.DataFrame = self.read() if read_on_init else None


    def read(self) -> None:
        '''
            Read the excel file

            .. code-block:: python
                # Only need if ReadExcel.read_on_init was left empty or set to False on init
                xlsx.read()
        '''
        self.data = pandas.read_excel(self.path)


    def get_sheet_names(self) -> list:
        '''
            Get the sheet names of the excel file

            .. code-block:: python
                sheets = xlsx.get_sheet_names()
        '''
        return pandas.ExcelFile(self.path).sheet_names


    def get_sheet_content(self,
                          sheet_name: str) -> pandas.DataFrame:
        '''
            Get the content of a sheet

            .. code-block:: python
                pd_frame = xlsx.get_sheet_content(sheet_name = "example_sheet_name")

            :param sheet_name: :class:`str` Sheet's name
        '''
        return pandas.read_excel(self.path, sheet_name)


    def get_sheet_content_to_list(self,
                                  sheet_name: str) -> list[list[any]]:
        '''
            Get the content of a sheet in list

            .. code-block:: python
                data_matrix = xlsx.get_sheet_content_to_list(sheet_name = "example_sheet_name")

            :param sheet_name: :class:`str` Sheet's name
        '''
        return pandas.read_excel(self.path,
                                 sheet_name,
                                 header = None).replace({numpy.nan: None}).values.tolist()



'''
    SMTP client module
'''

import smtplib
import ssl
from os.path import basename
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import COMMASPACE, formatdate
from email.mime.image import MIMEImage
from villog.log import Logger

class MailMan:
    '''
        SMTP client class
    '''
    __slots__: list[str] = ["smtp_server",
                            "smtp_login",
                            "smtp_port",
                            "smtp_password",
                            "name",
                            "do_logs",
                            "logger"]

    def __init__(self,
                 smtp_server: str,
                 smtp_login: str,
                 smtp_port: int,
                 smtp_password: str,
                 name: str,
                 do_logs: bool = True,
                 logger: Logger | None = None) -> None:
        '''
            SMTP client class

            .. code-block:: python
                demo_mail = MailMan(smtp_server = "smtp.example.com",
                                    smtp_login = "example@example.com",
                                    smtp_port = 1234,
                                    smtp_password = "example_password",
                                    name = "Ex Ample",
                                    do_logs = True,
                                    logger = Logger(file_path = "mail.log"))

            :param smtp_server: :class:`str` SMTP server
            :param smtp_login: :class:`str` SMTP login
            :param smtp_port: :class:`int` SMTP port
            :param smtp_password: :class:`str` SMTP password
            :param name: :class:`str` Name of the sender
            :param dp_logs: :class:`Optional(bool)` Let the mail manager log. Defaults to `True`
            :param logger: :class:`Union(Optional(Logger, None))` Logger class. Defaults to `None`
        '''
        self.smtp_server: str = smtp_server
        self.smtp_login: str = smtp_login
        self.smtp_port: str = smtp_port
        self.smtp_password: str = smtp_password
        self.name: str = name
        self.do_logs: bool = do_logs
        self.logger: Logger | None = logger if logger or do_logs else (Logger() if do_logs else None) # pylint: disable=line-too-long


    def __str__(self) -> str:
        return f"{self.smtp_login}@{self.smtp_server}:{self.smtp_port}"


    def __log(self,
              content: str) -> None:
        '''
            Log content
            
            :param content: :class:`str` Content to log
        '''
        if self.do_logs:
            self.logger.log(content)
        else:
            print(content)

    def send(self,
             subject: str,
             body: str,
             send_to: list[str],
             files: list[str] | None = None,
             images: list[str] | None = None) -> None:
        '''
            Send e-mail

            .. code-block:: python
                demo_mail.send(subject = "Example Mail",
                               body = "<html><body>Example mail content with image <img src='cid:logo'></body></html>",
                               send_to = ["recipient_1@example.com", "recipient_2@example.com"],
                               files = ["example.xlsx"],
                               images = ["logo.png"])

            :param subject: :class:`str` Subject of the email
            :param body: :class:`str` HTML body of the email
            :param send_to: :class:`list[str]` Recipients
            :param files: :class:`Optional(Union(list[str], None))` File(s) to attach. Defaults to `None`
            :param images: :class:`Optional(Union(list[str], None))` Image(s) to attach. Defaults to `None`
        ''' # pylint: disable = line-too-long
        self.__log(f"Sending mail {subject}")
        assert isinstance(send_to, list)
        msg: MIMEMultipart = MIMEMultipart()
        msg["From"] = f'{self.name} <{self.smtp_login}>'
        msg["To"] = COMMASPACE.join(send_to)
        msg["Date"] = formatdate(localtime = True)
        msg["Subject"] = subject
        context: ssl.SSLContext = ssl.create_default_context()

        msg.attach(MIMEText(body,
                            "html"))

        if images:
            for image in images:
                with open(image[0],
                          "rb") as file:
                    img: MIMEImage = MIMEImage(file.read())
                    img.add_header('Content-ID', "<" + image[1] + ">")
                    msg.attach(img)

        for f in files or []:
            with open(f, "rb") as file:
                part: MIMEApplication = MIMEApplication(
                    file.read(),
                    Name=basename(f)
                )
            part['Content-Disposition'] = f'attachment; filename="{basename(f)}"'
            msg.attach(part)
        smtp: smtplib.SMTP_SSL = smtplib.SMTP_SSL(self.smtp_server,
                                self.smtp_port,
                                context = context)
        smtp.login(self.smtp_login,
                   self.smtp_password)
        smtp.sendmail(self.smtp_login, 
                      send_to, msg.as_string())
        self.__log(f"Mail sent to {send_to}")
        smtp.close()



'''
    MsSQL module
'''

import os
import platform
from time import sleep
from dataclasses import dataclass
from ping3 import ping
import pyodbc
from villog.log import Logger
from villog.writexcel import WorkSheet, WorkBook

class VillSqlException(Exception):
    '''
        VillSQL's general exception
    '''
    def __init__(self,
                 message: str = "General exception") -> None:
        '''
            VillSQL's general exception

            :param message: :class:`str` Exception's message
        '''
        super().__init__(message)


class VillSqlPingException(Exception):
    '''
        VillSQL's ping exception
    '''
    def __init__(self,
                 message: str = "Can't reach the server") -> None:
        '''
            VillSQL's ping exception

            :param message: :class:`str` Exception's message
        '''
        super().__init__(message)


class VillSqlOdbcDriverException(Exception):
    '''
        VillSQL's ODBC driver exception
    '''
    def __init__(self,
                 message: str = "Can't get ODBC driver"):
        '''
            VillSQL's ODBC driver exception

            :param message: :class:`str` Exception's message
        '''
        super().__init__(message)


@dataclass(slots = True)
class SQLConfig:
    '''
        VillSQL configuration class
        Helps to store SQL configuration

        :param server: :class:`str`
        :param database: :class:`str`
        :param username: :class:`str`
        :param password: :class:`str`
    '''
    server: str
    database: str
    username: str
    password: str


class MsSQLClient:
    '''
        MsSQL client class
    '''

    __slots__: list[str] = ["server",
                            "database",
                            "username",
                            "password",
                            "allow_execute",
                            "logger",
                            "driver",
                            "connection",
                            "cursor"]

    def __init__(self,
                 server: str,
                 database: str,
                 username: str,
                 password: str,
                 is_trusted: bool = True,
                 allow_execute: bool = False,
                 logger: Logger | None = None) -> None:
        '''
            MsSQL client class

            :param server: :class:`str` Server name
            :param database: :class:`str` Database name
            :param username: :class:`str` Username
            :param password: :class:`str` Password
            :param is_trusted: :class:`Optional(bool)` Is the connection trusted? Defaults to `True`
            :param allow_execute: :class:`Optional(bool)` Does the connection allowed to make executions? Defaults to `False`
            :param logger: :class:`Optional(Union(Logger, None))` Logger class. Defaults to `None`
        ''' # pylint: disable=line-too-long
        self.server: str = server
        self.database: str = database
        self.allow_execute: bool = allow_execute
        self.logger: Logger | None = logger
        self.driver: str = self.__get_driver()
        if not self.driver:
            raise VillSqlOdbcDriverException("No ODBC driver found")
        if not self.__ping():
            raise VillSqlPingException(message = "Can't reach the server")
        try:
            self.connection: pyodbc.Connection = self.__connect(username, # pylint: disable=c-extension-no-member
                                                                password,
                                                                is_trusted)
            self.cursor: pyodbc.Connection.cursor = self.connection.cursor() # pylint: disable=c-extension-no-member
        except Exception as error:
            if self.__is_mac_arm():
                raise VillSqlException("If you got a pyodbc error and using a MacOS with ARM cpu, then try to install it with 'pip install --no-binary :all: pyodbc'") from error # pylint: disable=line-too-long
            raise error


    def __str__(self) -> str:
        return f"{self.database}@{self.server}"


    def __log(self,
              content: str) -> None:
        '''
            Log content

            :param content: :class:`str` Content to log
        '''
        if self.logger:
            self.logger.log(content)
        else:
            print(content)


    def __is_mac_os(self) -> bool:
        '''
            Check if the OS is Mac
        '''
        return 'darwin' in platform.system().lower()


    def __is_arm(self) -> bool:
        '''
            Check if the CPU is ARM
        '''
        return 'arm' in platform.machine().lower()


    def __is_mac_arm(self) -> bool:
        '''
            Check if the OS is Mac and the CPU is ARM
        '''
        return self.__is_mac_os() and self.__is_arm()


    def __ping(self,
               attempt: int = 5,
               wait: int = 10) -> bool:
        '''
            Ping the server

            :param attempt: :class:`int` How many time(s) should `self.__ping()` retry. Defaults to `5`
            :param wait: :class:`int` Wait between retries. Defaults to `10`
        ''' # pylint: disable=line-too-long
        if attempt != 0:
            response_time: float | None = ping(self.server)
            if response_time or response_time == 0:
                self.__log(f"{self.server} reached")
                return True
            else:
                self.__log(f"Can't reach {self.server}")
                attempt -= 1
                if attempt != 0:
                    self.__log(f"Retrying in {wait} seconds")
                    sleep(wait)
                    self.__ping(attempt, wait)
        self.__log(f"Can't reach {self.server}, no more attempts")
        return False


    def __get_driver(self) -> str | None:
        '''
            Get ODBC driver
        '''
        drivers: list[str] = pyodbc.drivers() # pylint: disable=c-extension-no-member
        for driver in drivers:
            if driver.startswith("ODBC Driver ") and driver.endswith(" for SQL Server"):
                return driver.replace("ODBC Driver ",
                                      "").replace(" for SQL Server",
                                                  "")
        if drivers:
            return drivers[0]
        return None


    def __connect(self,
                  username: str,
                  password: str,
                  is_trusted: bool = True) -> pyodbc.Connection: # pylint: disable=c-extension-no-member
        '''
            Connect to the server

            :param username: :class:`str` Username
            :param password: :class:`str` Password
            :param is_trusted: :class:`Optional(bool)` Is the connection trusted? Defaults to `True`
        '''
        connection_string: str = ""
        connection_string += 'DRIVER={ODBC Driver ' + self.driver
        connection_string += ' for SQL Server};SERVER=' + self.server
        connection_string += ';DATABASE=' + self.database
        connection_string += ';UID=' + username
        connection_string += ';PWD=' + password
        connection_string += ";TrustServerCertificate=" + ("yes" if is_trusted else "no") + ";"
        self.__log(f"Connecting to {self.database}@{self.server}")
        connection: pyodbc.Connection = pyodbc.connect(connection_string) # pylint: disable=c-extension-no-member
        self.__log(f"Connected to {self.database}@{self.server}")
        return connection


    def close(self) -> None:
        '''
            Close connection
        '''
        self.cursor.close()
        self.connection.close()
        self.__log(f"Connection to {self.server} closed")


    def execute(self,
                query: str,
                insert: list | None = None) -> None:
        '''
            Execute query

            .. code-block:: python
                mssql_client.execute(query = "select * from ?",
                                     insert = ("table_name"))

            :param query: :class:`str` query to run
            :param insert: :class:`Optional(Union(list, None))` tuple to insert into the query text
        '''
        if self.allow_execute:
            if insert is not None:
                self.cursor.execute(query, insert)
            else:
                self.cursor.execute(query)
            self.connection.commit()
            self.__log("Executed.")
        else:
            raise VillSqlException("Execution is not allowed, set 'allow_execute' to True")


    def select(self,
               query: str,
               insert: list | None = None) -> tuple | None:
        '''
            Select query

            .. code-block:: python
                columns, result = mssql_client.select(query = "select * from ?",
                                                      insert = ("table_name"))

            :param query: :class:`str` query to run
            :param insert: :class:`Optional(Union(list, None))` tuple to insert into the query text
        '''
        if insert is not None:
            self.cursor.execute(query, insert)
        else:
            self.cursor.execute(query)
        result = self.cursor.fetchall()
        columns = [description[0] for description in self.cursor.description]
        return columns, result


    def one_value_select(self,
                         query: str,
                         insert: list | None = None) -> any:
        '''
            Select one value
            (If the query returns more than one value, it will return the first one like `values[0][0]`)

            .. code-block:: python
                value = mssql_client.one_value_select(query = "select * from ?",
                                                      insert = ("table_name"))

            :param query: :class:`str` query to run
            :param insert: :class:`Optional(Union(list, None))` tuple to insert into the query text.
        ''' # pylint: disable=line-too-long
        _, result = self.select(query,
                                insert)
        return result[0][0] if result else None


class Table:
    '''
        Table class for Octopus 8
    '''

    __slots__ = ["columns",
                 "rows"]

    def __init__(self,
                 columns: list[str],
                 rows: list[list[any]]) -> None:
        '''
            Table class for Octopus

            :param columns: :class:`list[str]` Columns
            :param rows: :class:`list[list[any]]` Rows
        '''
        self.columns: list[str] = columns
        self.rows: list[list] = rows


    def __column_index(self,
                       column_name: str) -> int | None:
        '''
            Get column index

            :param column_name: :class:`str` Name of the column
        '''
        for index, column in enumerate(self.columns):
            if column.lower() == column_name.lower():
                return index
        return None


    def return_column(self,
                      column_name: str) -> list[str] | None:
        '''
            Return the given column if exists

            .. code-block:: python
                column_1 = demo_table.return_column(column_name = "column_1")
            
            :param column_name: :class:`str` Column name
        '''
        if isinstance(column_name, list):
            raise VillSqlException("For multiple columns use 'return_columns'")
        column_index: int | None = self.__column_index(column_name)
        if column_index is not None:
            return [row[column_index] for row in self.rows]
        return None


    def return_columns(self,
                       column_names: list[str]) -> list[list] | None:
        '''
            Return the given columns if exists

            .. code-block:: python
                column_1_2_matrix = demo_table.return_columns(column_names = ["column_1", "column_2"])

            :param column_names: :class:`list[str]` Column names
        ''' # pylint: disable=line-too-long
        if not isinstance(column_names,
                          list):
            if isinstance(column_names,
                          str):
                raise VillSqlException("For single column use 'return_column'")
            raise VillSqlException("Column names must be a list")
        new_rows: list[list] = []
        for row in self.rows:
            new_row: list = []
            for column_name in column_names:
                column_index: int = self.__column_index(column_name)
                if column_index is not None:
                    new_row.append(row[column_index])
            new_rows.append(new_row)
        return new_rows


    def set_filter(self,
                   column_names: list[str]) -> None:
        '''
            Filter table by column names

            .. code-block:: python
                demo_table.set_filter(column_names = ["column_1", "column_2"])

            :param column_names: :class:`str` Column names
        '''
        self.rows = self.return_columns(column_names)
        self.columns = column_names


    def export_to_excel(self,
                        path: str) -> str:
        '''
            Export table to excel

            .. code-block:: python
                demo_table.export_to_excel(path = "demo_table.xlsx")
            
            :param path: :class:`str` Path to create excel
        '''
        WorkBook(name = "Book1",
                 sheets = WorkSheet(name = "Sheet1",
                                    header = self.columns,
                                    data = self.rows)).xlsx_create(file_path = path)


class VillSQL:
    '''
        Octopus 8 class
    '''

    __slots__: list[str] = ["__do_logs",
                            "__logger",
                            "__client",
                            "__row_limit",
                            "__tables",
                            "__allow_execute"]

    def __init__(self,
                 sql_config: SQLConfig | None = None,
                 server: str | None = None,
                 database: str | None = None,
                 username: str | None = None,
                 password: str | None = None,
                 is_server_trusted: bool = True,
                 do_logs: bool = True,
                 logger: Logger | None = None,
                 row_limit: int | None = None,
                 do_table_fetch: bool = False,
                 allow_execute: bool = False) -> None:
        '''
            Octopus 8 class

            :param sql_config: :class:`Optional(Union(SQLConfig, None))` SQL config for easier configuration. Defaults to `None`
            :param server: :class:`Optional(Union(str, None))` Server name. Defaults to `None`
            :param database: :class:`Optional(Union(str, None))` Database name. Defaults to `None`
            :param username: :class:`Optional(Union(str, None))` Username. Defaults to `None`
            :param password: :class:`Optional(Union(str, None))` Password. Defaults to `None`
            :param is_server_trusted: :class:`Optional(bool)` Is the connection trusted? Defaults to `True`
            :param do_logs: :class:`Optional(bool)` Let the class make logs. Defaults to `True`
            :param logger: :class:`Optional(Union(Logger, None))` Logger to make logs by. Defaults to `None`
            :param row_limit: :class:`Union(Optional(int, None))` Row limit of selects. Defaults to `None`
            :param do_table_fetch: :class:`Optional(bool)` Fetch tables to variables. Defaults to `False`
            :param allow_execute: :class:`Optional(bool)` Does the connection allowed to make executions? Defaults to `False`
        ''' # pylint: disable=line-too-long
        self.__do_logs: bool = do_logs
        if do_logs:
            self.__logger: Logger = logger or Logger(file_path = os.path.join(os.getcwd(),
                                                                              "octopus.log"))
        server_c: str = sql_config.server if sql_config else server
        database_c: str = sql_config.database if sql_config else database
        username_c: str = sql_config.username if sql_config else username
        password_c: str = sql_config.password if sql_config else password
        self.__client: MsSQLClient = MsSQLClient(server = server_c,
                                                 database = database_c,
                                                 username = username_c,
                                                 password = password_c,
                                                 is_trusted = is_server_trusted,
                                                 logger = logger,
                                                 allow_execute = allow_execute)
        del server_c, database_c, username_c, password_c
        self.__row_limit: int | None = row_limit
        self.__tables: list[str] = self.__get_tables()

        if do_table_fetch:
            for table in self.__tables:
                if table.lower() != "table":
                    setattr(self,
                            f"get_table_{table.lower()}",
                            lambda table = table,
                            raw_filter = "",
                            order_by = None,
                            **kfilter:
                            self.get_table(table,
                                           raw_filter,
                                           order_by,
                                           **kfilter))
        self.__allow_execute: bool = allow_execute


    def __str__(self) -> str:
        return str(self.__client)


    def __log(self,
              content: str) -> None:
        '''
            Log content
        
            :param content: :class:`str` Content to log
        '''
        if self.__do_logs:
            self.__logger.log(content)
        else:
            print(content)


    def __get_row_limit(self) -> str:
        '''
            Get row limit for query
        '''
        if self.__row_limit and self.__row_limit > 0:
            return f" top {str(self.__row_limit)}"
        return ""


    def set_row_limit(self,
                      row_limit: int) -> None:
        '''
            Set row limit (like using `select top n`)
        
            .. code-block:: python
                demo_sql.set_row_limit(row_limit = 100)

            :param row_limit: :class:`int` Set row limit for selects
        '''
        self.__row_limit = row_limit


    def __get_tables(self) -> list[str]:
        '''
            Get tables from database
        '''
        print("Getting tables")
        _, results = self.__client.select("select name from sys.tables")
        tables = [result[0] for result in results]
        if not tables:
            raise VillSqlException("No tables found")
        return tables


    def __is_table(self,
                   table: str) -> bool:
        '''
            Check if table exists

            :param table: :class:`str` Table to check
        '''
        for tbl in self.__tables:
            if tbl.lower() == table.lower():
                self.__log(f"Table '{table}' found")
                return True
        self.__log(f"Table '{table}' not found")
        return False


    def __kfilter_to_query(self,
                           kfilter_item: list) -> str:
        '''
            Convert kwargs filter to query
        
            :param kfilter_item: :class:`list` kwargs filter item
        '''
        if isinstance(kfilter_item[1],
                      str):
            if kfilter_item[1].lower() in ("null", "is null"):
                return f"{kfilter_item[0]} is null"
            if kfilter_item[1].lower() in ("not null", "is not null"):
                return f"{kfilter_item[0]} is not null"
            return f"{kfilter_item[0]} = '{kfilter_item[1]}'"
        if isinstance(kfilter_item[1],
                      int):
            return f"{kfilter_item[0]} = {kfilter_item[1]}"


    def get_table(self,
                  table: str,
                  raw_filter: str = "",
                  order_by: list[tuple] | None = None,
                  **kfilter) -> Table:
        '''
            Selects a table from the database and returns it as a Table object

            .. code-block:: python
                table_result = demo_client.get_table(table = "TABLE_NAME",
                                                     raw_filter = "column_1 like '%xyz%'",
                                                     order_by = [("column_1", "ASC"),
                                                                 ("column_2", "DESC")],
                                                     column_2 = 1)

            :param table: :class:`str` Table's name
            :param raw_filter: :class:`Optional(str)` Raw filter string to where clause. Defaults to `""`
            :param order_by: :class:`Optional(Union(list[tuple], None))` Order by clause. Defaults to `None`
            :param **kfilter: :class:`**kwargs` Kwargs filter adter the where caluse. Default to `{}`
        ''' # pylint: disable=line-too-long
        if self.__is_table(table):
            query: str = f"select {self.__get_row_limit()} * from {table} with (nolock) where 1 = 1"
            if raw_filter:
                query += f" and {raw_filter}"
            if kfilter:
                for kfilter_item in kfilter.items():
                    query += f" and {self.__kfilter_to_query(kfilter_item)}"
            if order_by:
                query += " order by "
                for i, order in enumerate(order_by):
                    query += f"{order[0]} {order[1]}{'' if len(order_by) - 1 == i else ', '} "
            self.__log(f"Executing: {(query)}")
            columns, result = self.__client.select(query)
            if result:
                return Table(columns, result)
        return None


    def custom_query(self,
                     query: str,
                     inserter: list[any] | None = None) -> tuple | None:
        '''
            Select query

            .. code-block:: python
                columns, result = demo_client.custom_query(query = "select * from ?",
                                                           inserter = ("table_name"))

            :param query: :class:`str` query to run
            :param inserter: :class:`Optional(Union(list, None))` tuple to insert into the query text
        ''' # pylint: disable=line-too-long
        if self.__allow_execute:
            return self.__client.select(query,
                                        inserter)
        raise VillSqlException("Custom query is only allowed if 'allow_execute' is set to True")


    def custom_query_to_table(self,
                              query: str,
                              inserter: list[any] | None = None) -> Table:
        '''
            Query to :class:`Table`

            .. code-block:: python
                table_result = demo_client.custom_query_to_table(query = "select * from ?",
                                                                 inserter = ("table_name"))

            :param query: :class:`str` query to run
            :param inserter: :class:`Optional(Union(list, None))` tuple to insert into the query text
        ''' # pylint: disable=line-too-long
        columns, result = self.custom_query(query,
                                            inserter)
        if result:
            return Table(columns, result)
        return None


    def custom_query_only_columns(self,
                                  query: str,
                                  inserter: list[any] | None = None) -> list[str] | None:
        '''
            Query only columns

            .. code-block:: python
                columns = demo_client.custom_query_only_columns(query = "select * from ?",
                                                                inserter = ("table_name"))

            :param query: :class:`str` query to run
            :param inserter: :class:`Optional(Union(list, None))` tuple to insert into the query text
        ''' # pylint: disable=line-too-long
        columns, _ = self.__client.select(query,
                                          inserter)
        return columns


    def custom_query_only_values(self,
                                 query: str,
                                 inserter: list[any] | None = None) -> list[list] | None:
        '''
            Query only values

            .. code-block:: python
                values = demo_client.custom_query_only_values(query = "select * from ?",
                                                              inserter = ("table_name"))

            :param query: :class:`str` query to run
            :param inserter: :class:`Optional(Union(list, None))` tuple to insert into the query text
        ''' # pylint: disable=line-too-long
        _, result = self.__client.select(query,
                                         inserter)
        return result


    def execute(self,
                query: str,
                insert: list | None = None) -> None:
        '''
            Execute query

            .. code-block:: python
                mssql_client.execute(query = "select * from ?",
                                     insert = ("table_name"))

            :param query: :class:`str` query to run
            :param insert: :class:`Optional(Union(list, None))` tuple to insert into the query text
        ''' # pylint: disable=line-too-long
        if self.__allow_execute:
            self.__client.execute(query,
                                  insert)
        else:
            raise VillSqlException("Execution is not allowed, set 'allow_execute' to True")


    def __read_file(self,
                    path: str,
                    encoding: str = "utf-8-sig") -> str:
        '''
            Read file content

            :param path: :class:`str` Path to the file
            :param encoding: :class:`str` Encoding of the file. Defaults to `utf-8-sig`
        '''
        if not os.path.exists(path):
            raise VillSqlException(f"File not found: {path}")
        with open(file = path,
                  mode = "r",
                  encoding = encoding) as file:
            return file.read()


    def execute_file(self,
                     path: str,
                     encoding: str = "utf-8-sig") -> None:
        '''
            Execute file content

            .. code-block:: python
                mssql_client.execute_file(path = "example.sql")
            
            :param path: :class:`str` Path to the file
            :param encoding: :class:`str` Encoding of the file. Defaults to `utf-8-sig`
        '''
        return self.custom_query(self.__read_file(path,
                                                  encoding))


    def execute_file_to_table(self,
                              path: str,
                              encoding: str = "utf-8-sig") -> Table:
        '''
            Execute file content and return as :class:`Table`

            .. code-block:: python
                demo_table = mssql_client.execute_file_to_table(path = "example.sql")

            :param path: :class:`str` Path to the file
            :param encoding: :class:`str` Encoding of the file. Defaults to `utf-8-sig`
        '''
        return self.custom_query_to_table(self.__read_file(path,
                                                           encoding))


    def close(self) -> None:
        '''
            Close connection
        '''
        self.__client.close()



'''
    This module is responsible for generating PDF files from HTML and CSS files.
'''

try:
    from weasyprint import HTML, CSS

    def generate(html_string: str,
                 output_path: str,
                 css_string: str | None = None) -> None:
        '''
            Generate a PDF file from a HTML and CSS file.

            .. code-block:: python
                generate(html_string = "<html><body>test</body></html>",
                         output_path = "example.pdf,
                         css_string = "body { color: red; }")

            :param html_string: :class:`str` HTML string
            :param output_path: :class:`str` Output path for the .pdf file
            :param css_string: :class:`Optional(Union(str, None))` CSS string. Defaults to `None`
        '''
        HTML(string = html_string).write_pdf(output_path,
                                            stylesheets = [CSS(string = css_string or "")]) # pylint: disable = line-too-long
except Exception as e: # pylint: disable=broad-exception-caught
    print(e)



'''
    Package setup
'''

from pathlib import Path
from setuptools import setup, find_packages

def parse_requirements(filename: str) -> list:
    '''
        Parse a requirements file returning a list of requirements.

        Args:
            filename (str): The path to the requirements file.
    '''
    requirements_path: str = Path(__file__).parent / filename
    with requirements_path.open() as f:
        return [line.strip() for line in f if line.strip() and not line.startswith("#")]

this_directory: str = Path(__file__).parent
long_description: str = (this_directory / "README.md").read_text()


setup(name = "villog",
      version = "0.3.1",
      description = "A simple python utility tool for your everyday projects.",
      author = "Krisztián Villers",
      packages = find_packages(),
      long_description = long_description,
      long_description_content_type = "text/markdown",
      install_requires = parse_requirements("requirements.txt"))
